import pandas as pd
import pickle
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import math
import os
from os.path import join
import sys
from aboss.reference_matrix import calculate_posCount, cdrh3,frame_canon
from aboss.IMGT_germline_reference import adding_zeros


def calculateEntropy( control_frame_prop, length_of_frame ):

    entropy = 0
    average_AAUsage = {}

    for ele in control_frame_prop:
        average_AAUsage[ele] = control_frame_prop[ele] / float(length_of_frame)
        if average_AAUsage[ele] > 0:
            entropy += - average_AAUsage[ele]*math.log(average_AAUsage[ele], 2)
    return entropy, average_AAUsage

def plotBarplot(df, color, frame, description):

    fig, axis = plt.subplots(figsize=(8,8) )
    df.plot.bar(title ="{0} {1}, ABOSS filtered data".format(frame, description), ax=axis, legend = False, color=color)
    plt.savefig("Graphs/{0} {1} ABOSS filtered data".format(frame, description), dpi= 300)
    plt.clf()

def plotAAplot(df, frame):

    # Color dict for amino acids
    colors_dict = dict([("A", "#a2a0a5"),("C", "#fff319"),("D","#22a56f"),("E","#19ff34"),("F","#031dc1"),("G","#35ffee"),("H","#6e728c"),("I","#057200"), ("K","#720070"),("L","#914b00"),("M","#847806"),("N","#f24d76"),("P","#fcc7ed"),("Q","#cc0c0c"),("R","#2cc124"),("S","#c124c1"),("T","#ff6600"),("V","#c6e833"),("W","#ff0000"),("Y","#6fa522")])
    my_colors = [colors_dict[x] for x in list(df)]

    # plotting
    fig, axis = plt.subplots(figsize=(10,8))
    df.plot.bar(stacked=True, color = my_colors, ax =axis, title ="{0} proportions, ABOSS filtered data".format(frame), legend='reverse').legend(bbox_to_anchor=(0.99, 1.02))
    plt.savefig("Graphs/{0} proportion ABOSS filtered data".format(frame),dpi= 300)
    plt.clf()

def orderIMGT(cols, frame):
    if frame not in ["cdrh3","cdrl3"]:
        new_cols = frame_canon(cols)
    else:
        new_cols = cdrh3(cols)	
    return new_cols

class PlotGraps(object):
	"""Plotting three types of graphs here:
		1) Amino acid usage by position in ABOSS filtered Ig-seq Data
		2) Entropy in antibody regions in ABOSS filtered Ig-seq Data
		3) Counts of amino acid entries at positions in ABOSS filtered Ig-seq Data
	Inputs:
		@frame: Analysed antibody region such as cdrh3
		@regions_aaCount: Counts of amino acid residues at positions in ABOSS filtered Ig-seq Data 
	"""
        def __init__(self, frame, regions_aaCount):
            self.frame = frame

            # with 0 flagged residue/positions
            self.regions_aaCount = regions_aaCount[0]

            self.entropy_dict = {self.frame:{}}
            self.createGraphsDir()
            self.prepareData()

        def createGraphsDir(self):
            if not os.path.isdir("Graphs"):
                    os.makedirs("Graphs")

        def control_bar(self):
                """ Function to calcualte the average amino acid distribution and entropies in antibody regions
                Return:
                        @average_AAUsage: Average amino acid usage in a given antibody region
                        @entropy_contr:   Average entropy of positions in a given antibody region
                """
                entropy_contr = {}
                control_frame_prop = {self.frame:{}}
                average_AAUsage = {self.frame:{}} 
                for pos in self.regions_aaCount:
                        for ele in self.regions_aaCount[pos]:
                            if ele not in control_frame_prop[self.frame]:
                                control_frame_prop[self.frame][ele] = 0
                            control_frame_prop[self.frame][ele] += self.regions_aaCount[pos][ele]

                length_of_frame = sum([ self.pos_Count[pos] for pos in self.pos_Count])
                                
                entropy, average_AAUsage_f =  calculateEntropy( control_frame_prop[self.frame], length_of_frame )
                entropy_contr[self.frame] = entropy
                average_AAUsage[self.frame] = average_AAUsage_f
                return average_AAUsage, entropy_contr

        def prepareData(self):

            self.regions_aaCount = adding_zeros(self.regions_aaCount)[self.frame]
            self.pos_Count = calculate_posCount(self.regions_aaCount)
            weighted_contr, _ = self.control_bar()

            # Here we calculate entropy by position
            for pos in self.regions_aaCount:
                entropy, average_AAUsage_f = calculateEntropy( self.regions_aaCount[pos], self.pos_Count[pos])
                self.entropy_dict[self.frame][pos] = entropy
                self.regions_aaCount[pos] = average_AAUsage_f

            assert weighted_contr != {self.frame: {}},"Input file didn't contain sequence of antibody region"
            
            # DataFrame of Position Entropy
            df_entropy = pd.DataFrame(self.entropy_dict[self.frame], index = [0]).fillna(0)  
            
            # DataFrame of Amino Acid Distribution
            df_aa_proportion = pd.DataFrame(self.regions_aaCount).fillna(0)	     

            # DataFrame of Entries by position
            df_pos_len = pd.DataFrame(self.pos_Count, index=[0]).fillna(0) 	      

            # Ordering IMGT positions
            new_cols = orderIMGT( df_aa_proportion.columns.tolist(), self.frame)
            
            # plotting entropy
            entropy_complete_plot = df_entropy[new_cols].transpose()
            plotBarplot(entropy_complete_plot, "red", self.frame, "entropy")

            # plotting entries
            df_pos_len = df_pos_len[new_cols].transpose()
            plotBarplot( df_pos_len, "green", self.frame, "entries")

            # AA distribution by position plot
            df_aa_proportion = df_aa_proportion[new_cols].transpose()
            
            # Apending weighted control bar
            averaged_aa_control_bar = pd.DataFrame(weighted_contr).fillna(0).transpose() 
            aa_proportion_complete_plot = df_aa_proportion.append(averaged_aa_control_bar)

            plotAAplot(aa_proportion_complete_plot , self.frame)
            return;
if __name__ == '__main__':
	pass
