import pandas as pd
import cPickle as pickle
import math
import os
import sys
from aboss.run_ANARCI_parsing import FW1, FW3  


def calculate_posCount(region_aaCount):
	"""Function that calculates the total number of amino acids at positions in antibody regions
	Inputs:	
		@region_aaCount: Counts of amino acids at positions in a given antibody region (e.g. framework 1)
	Return:
		@pos_Count: The total number of amino acids at positions
	"""
	pos_Count = {}

	for pos in region_aaCount:
		for ele in region_aaCount[pos]:
			if pos not in pos_Count:
				pos_Count[pos] = 0
			pos_Count[pos] += region_aaCount[pos][ele]

	return pos_Count

def noise_removal(error, IMGT_reference, pos_amino_acids, pos_Count, framework):
	"""Function that creates a reference matrix for an antibody region
	Inputs:
		@error:     Residue error rate
		@pos_amino: Amino acid counts by position in an antibody region
		@pos_Count: Total number of amino acids at a position

	Return:
		@amino_acid_by_pos: Matrix of counts of amino acids at positions if their proportions are above the residue error rate

	"""
	amino_acid_by_pos = {}
	for position in pos_amino_acids:
		if position not in amino_acid_by_pos:
			amino_acid_by_pos[position] = {}
		for amino in pos_amino_acids[position]:
                    amino_pos_proportion = float(pos_amino_acids[position][amino]) / pos_Count[position]

                    if position in IMGT_reference[framework]:
                            if IMGT_reference[framework][position][amino] == 0:
                                    if amino_pos_proportion <= error:
                                            amino_acid_by_pos[position][amino] = 0
                                    else:
                                            amino_acid_by_pos[position][amino] = pos_amino_acids[position][amino]
                            else:
                                    amino_acid_by_pos[position][amino] = pos_amino_acids[position][amino]
                    else:
                            if amino_pos_proportion <= error:
                                     amino_acid_by_pos[position][amino] = 0
                            else:
                                    amino_acid_by_pos[position][amino] = pos_amino_acids[position][amino]
	return amino_acid_by_pos

def calculateResidueErrorRate(regions_aaCount, chain):
	"""Function calculate the residue error rate for Ig-seq Data
	Inputs:
		@regions_aaCount: Counts of amino acids at antibody positions in Ig-seq data
		@chain: Igseq data antibody chain
	Return:
		@error: The residue error rate value
	"""

	error = 0

        error_fw1 = sorted(regions_aaCount[FW1[chain]]["23"].values(), key=lambda x: x, reverse=True)
        error_fw1 = [float(x) / sum(error_fw1) for x in error_fw1]
        error_fw3 = sorted(regions_aaCount[FW3[chain]]["104"].values(), key=lambda x: x, reverse=True)	
        error_fw3 = [float(x) / sum(error_fw3) for x in error_fw3]

	if len(error_fw1) == 1 and len(error_fw3) == 1 :
		error = 0
	else:
		if len(error_fw1) > 1 and len(error_fw3) > 1:		
                    error = max(error_fw1[1], error_fw3[1])

		elif len(error_fw1) == 1:
			error = error_fw3[1]
		else:
			error = error_fw1[1]
	return error
			
def frame_canon(cols):
	"""
        Reordering canonical CDRs and frameworks according to IMGT definition
	"""
	
	integer_list =  [int(x) for x in cols]
	sorted_integer_list = sorted(integer_list,key=int)
	sorted_string_list = [str(x) for x in sorted_integer_list]
	return sorted_string_list

def cdrh3(cols):
	"""
        Reordering CDRH3 or CDRL3 according to IMGT definition
	"""
	integer_list = []
	pos_111 = []
	pos_112 = []
	sorted_string_list = []
	for pos in cols:
            try:
                integer_list.append(int(pos))
                sorted_integer_list = sorted(integer_list,key=int)
                sorted_string_list = [str(x) for x in sorted_integer_list]
            except:
                if "111" in pos:
                        pos_111.append(pos)
                elif "112" in pos:
                        pos_112.append(pos)
	if "112A" in pos_112:
		pos_111_112_list = sorted(pos_111) + sorted(pos_112)[::-1]
		sorted_string_list = sorted_string_list[:sorted_string_list.index("111")+1] + pos_111_112_list + sorted_string_list[sorted_string_list.index("111")+1:]
	return sorted_string_list					

class CreateRefMatrix(object):
	"""Class that creates a reference matrix for Ig-seq data.
	Inputs:
		@chain = Igseq data antibody chain
		@frame = antibody region (can be framework or cdr)
		@regions_aaCount = Counts of amino acids at antibody positions in Ig-seq data
                @residueErrorRate = precalculated residue error
	"""

        def __init__(self, chain, frame, regions_aaCount, IMGT_reference, residueErrorRate):
            self.chain = chain
            self.frame = frame
            self.regions_aaCount = regions_aaCount
            self.IMGT_reference = IMGT_reference
            self.residueErrorRate = residueErrorRate
            self.createTempDirs()

        def createTempDirs(self):
            if os.path.isdir("Reference_matrix") != True:
                    os.makedirs("Reference_matrix")
            if os.path.isdir("Error") != True:
                    os.makedirs("Error")

        def convertToProportion(self, counting_aa, pos_Count):

            for pos in counting_aa:
                    entropy = 0
                    for ele in counting_aa[pos]:
                             counting_aa[pos][ele] = float(counting_aa[pos][ele]) / pos_Count[pos]
            return counting_aa

        def create(self):

            # Calculating total number of entries at positions
            pos_Count = calculate_posCount(self.regions_aaCount[self.frame])

            # Checking if residues/positions are above the residue error rate
            counting_aa = noise_removal(self.residueErrorRate, self.IMGT_reference, self.regions_aaCount[self.frame], pos_Count, self.frame) 

            # Recalculating number of entries at positions  
            pos_Count = calculate_posCount(counting_aa)  

            # Converting residue/positions into proportions
            counting_aa_prop = self.convertToProportion(counting_aa, pos_Count)  

            # Pandas DataFrame of amino acid distribution by position                 
            df_aa_proportion = pd.DataFrame(counting_aa_prop).fillna(0) 

            # Pandas DataFrame to check the number of entries at position
            pos_hist = pd.DataFrame(pos_Count, index = [0]).fillna(0)

            # Deleting positions with fewer than 20 entries
            colsToDel = list(pos_hist.T[pos_hist.T[0] <= 20].index)
            if colsToDel != []:
                df_aa_proportion.drop(columns=colsToDel, inplace=True)
                print "Positions with fewer than 20 entries: {0}".format(colsToDel)       

            # Re-ordering IMGT numbering
            cols = df_aa_proportion.columns.tolist()
            if self.frame not in ["cdrh3","cdrl3"]:
                    new_cols = frame_canon(cols)
            else:
                    new_cols = cdrh3(cols)		

            # Saving outputs
            with open("Reference_matrix/reference_matrix_{0}.pkl".format(self.frame), "wb") as y:
                    pickle.dump(df_aa_proportion[new_cols].transpose(),  y)

            with open("Error/ig_seq_residue_error_rate.txt", "w") as fi:
                    fi.write("\n")
                    fi.write("The residues error rate for this Ig-seq dataset is {0} \n".format(self.residueErrorRate))
            return; 

if __name__ == '__main__':

	pass
